@@ -1,7 +1,6 @@
 from __future__ import with_statement
 from distutils.version import StrictVersion
 from itertools import chain
-from select import select
 import os
 import socket
 import sys
@@ -16,7 +15,8 @@ except ImportError:
 
 from redis._compat import (b, xrange, imap, byte_to_chr, unicode, bytes, long,
                            BytesIO, nativestr, basestring, iteritems,
-                           LifoQueue, Empty, Full, urlparse, parse_qs)
+                           LifoQueue, Empty, Full, urlparse, parse_qs,
+                           recv, recv_into, select, unquote)
 from redis.exceptions import (
     RedisError,
     ConnectionError,
@@ -36,17 +36,28 @@ if HIREDIS_AVAILABLE:
     hiredis_version = StrictVersion(hiredis.__version__)
     HIREDIS_SUPPORTS_CALLABLE_ERRORS = \
         hiredis_version >= StrictVersion('0.1.3')
+    HIREDIS_SUPPORTS_BYTE_BUFFER = \
+        hiredis_version >= StrictVersion('0.1.4')
 
-    if not HIREDIS_SUPPORTS_CALLABLE_ERRORS:
-        msg = ("redis-py works best with hiredis >= 0.1.3. You're running "
+    if not HIREDIS_SUPPORTS_BYTE_BUFFER:
+        msg = ("redis-py works best with hiredis >= 0.1.4. You're running "
                "hiredis %s. Please consider upgrading." % hiredis.__version__)
         warnings.warn(msg)
 
+    HIREDIS_USE_BYTE_BUFFER = True
+    # only use byte buffer if hiredis supports it and the Python version
+    # is >= 2.7
+    if not HIREDIS_SUPPORTS_BYTE_BUFFER or (
+            sys.version_info[0] == 2 and sys.version_info[1] < 7):
+        HIREDIS_USE_BYTE_BUFFER = False
+
 SYM_STAR = b('*')
 SYM_DOLLAR = b('$')
 SYM_CRLF = b('\r\n')
 SYM_EMPTY = b('')
 
+SERVER_CLOSED_CONNECTION_ERROR = "Connection closed by server."
+
 
 class Token(object):
     """
@@ -54,10 +65,27 @@ class Token(object):
     hard-coded arguments are wrapped in this class so we know not to apply
     and encoding rules on them.
     """
+
+    _cache = {}
+
+    @classmethod
+    def get_token(cls, value):
+        "Gets a cached token object or creates a new one if not already cached"
+
+        # Use try/except because after running for a short time most tokens
+        # should already be cached
+        try:
+            return cls._cache[value]
+        except KeyError:
+            token = Token(value)
+            cls._cache[value] = token
+            return token
+
     def __init__(self, value):
         if isinstance(value, Token):
             value = value.value
         self.value = value
+        self.encoded_value = b(value)
 
     def __repr__(self):
         return self.value
@@ -66,9 +94,43 @@ class Token(object):
         return self.value
 
 
+class Encoder(object):
+    "Encode strings to bytes and decode bytes to strings"
+
+    def __init__(self, encoding, encoding_errors, decode_responses):
+        self.encoding = encoding
+        self.encoding_errors = encoding_errors
+        self.decode_responses = decode_responses
+
+    def encode(self, value):
+        "Return a bytestring representation of the value"
+        if isinstance(value, Token):
+            return value.encoded_value
+        elif isinstance(value, bytes):
+            return value
+        elif isinstance(value, (int, long)):
+            value = b(str(value))
+        elif isinstance(value, float):
+            value = b(repr(value))
+        elif not isinstance(value, basestring):
+            # an object we don't know how to deal with. default to unicode()
+            value = unicode(value)
+        if isinstance(value, unicode):
+            value = value.encode(self.encoding, self.encoding_errors)
+        return value
+
+    def decode(self, value, force=False):
+        "Return a unicode string from the byte representation"
+        if (self.decode_responses or force) and isinstance(value, bytes):
+            value = value.decode(self.encoding, self.encoding_errors)
+        return value
+
+
 class BaseParser(object):
     EXCEPTION_CLASSES = {
-        'ERR': ResponseError,
+        'ERR': {
+            'max number of clients reached': ConnectionError
+        },
         'EXECABORT': ExecAbortError,
         'LOADING': BusyLoadingError,
         'NOSCRIPT': NoScriptError,
@@ -80,7 +142,10 @@ class BaseParser(object):
         error_code = response.split(' ')[0]
         if error_code in self.EXCEPTION_CLASSES:
             response = response[len(error_code) + 1:]
-            return self.EXCEPTION_CLASSES[error_code](response)
+            exception_class = self.EXCEPTION_CLASSES[error_code]
+            if isinstance(exception_class, dict):
+                exception_class = exception_class.get(response, ResponseError)
+            return exception_class(response)
         return ResponseError(response)
 
 
@@ -106,10 +171,10 @@ class SocketBuffer(object):
 
         try:
             while True:
-                data = self._sock.recv(socket_read_size)
+                data = recv(self._sock, socket_read_size)
                 # an empty string indicates the server shutdown the socket
-                if isinstance(data, str) and len(data) == 0:
-                    raise socket.error("Connection closed by remote server.")
+                if isinstance(data, bytes) and len(data) == 0:
+                    raise socket.error(SERVER_CLOSED_CONNECTION_ERROR)
                 buf.write(data)
                 data_length = len(data)
                 self.bytes_written += data_length
@@ -168,18 +233,25 @@ class SocketBuffer(object):
         self.bytes_read = 0
 
     def close(self):
-        self.purge()
-        self._buffer.close()
+        try:
+            self.purge()
+            self._buffer.close()
+        except:
+            # issue #633 suggests the purge/close somehow raised a
+            # BadFileDescriptor error. Perhaps the client ran out of
+            # memory or something else? It's probably OK to ignore
+            # any error being raised from purge/close since we're
+            # removing the reference to the instance below.
+            pass
         self._buffer = None
         self._sock = None
 
 
 class PythonParser(BaseParser):
     "Plain Python parsing class"
-    encoding = None
-
     def __init__(self, socket_read_size):
         self.socket_read_size = socket_read_size
+        self.encoder = None
         self._sock = None
         self._buffer = None
 
@@ -193,8 +265,7 @@ class PythonParser(BaseParser):
         "Called when the socket connects"
         self._sock = connection._sock
         self._buffer = SocketBuffer(self._sock, self.socket_read_size)
-        if connection.decode_responses:
-            self.encoding = connection.encoding
+        self.encoder = connection.encoder
 
     def on_disconnect(self):
         "Called when the socket disconnects"
@@ -204,7 +275,7 @@ class PythonParser(BaseParser):
         if self._buffer is not None:
             self._buffer.close()
             self._buffer = None
-        self.encoding = None
+        self.encoder = None
 
     def can_read(self):
         return self._buffer and bool(self._buffer.length)
@@ -212,7 +283,7 @@ class PythonParser(BaseParser):
     def read_response(self):
         response = self._buffer.readline()
         if not response:
-            raise ConnectionError("Socket closed on remote end")
+            raise ConnectionError(SERVER_CLOSED_CONNECTION_ERROR)
 
         byte, response = byte_to_chr(response[0]), response[1:]
 
@@ -251,8 +322,8 @@ class PythonParser(BaseParser):
             if length == -1:
                 return None
             response = [self.read_response() for i in xrange(length)]
-        if isinstance(response, bytes) and self.encoding:
-            response = response.decode(self.encoding)
+        if isinstance(response, bytes):
+            response = self.encoder.decode(response)
         return response
 
 
@@ -263,6 +334,9 @@ class HiredisParser(BaseParser):
             raise RedisError("Hiredis is not installed")
         self.socket_read_size = socket_read_size
 
+        if HIREDIS_USE_BYTE_BUFFER:
+            self._buffer = bytearray(socket_read_size)
+
     def __del__(self):
         try:
             self.on_disconnect()
@@ -280,8 +354,8 @@ class HiredisParser(BaseParser):
         if not HIREDIS_SUPPORTS_CALLABLE_ERRORS:
             kwargs['replyError'] = ResponseError
 
-        if connection.decode_responses:
-            kwargs['encoding'] = connection.encoding
+        if connection.encoder.decode_responses:
+            kwargs['encoding'] = connection.encoder.encoding
         self._reader = hiredis.Reader(**kwargs)
         self._next_response = False
 
@@ -292,7 +366,7 @@ class HiredisParser(BaseParser):
 
     def can_read(self):
         if not self._reader:
-            raise ConnectionError("Socket closed on remote end")
+            raise ConnectionError(SERVER_CLOSED_CONNECTION_ERROR)
 
         if self._next_response is False:
             self._next_response = self._reader.gets()
@@ -300,7 +374,7 @@ class HiredisParser(BaseParser):
 
     def read_response(self):
         if not self._reader:
-            raise ConnectionError("Socket closed on remote end")
+            raise ConnectionError(SERVER_CLOSED_CONNECTION_ERROR)
 
         # _next_response might be cached from a can_read() call
         if self._next_response is not False:
@@ -312,23 +386,25 @@ class HiredisParser(BaseParser):
         socket_read_size = self.socket_read_size
         while response is False:
             try:
-                buffer = self._sock.recv(socket_read_size)
-                # an empty string indicates the server shutdown the socket
-                if isinstance(buffer, str) and len(buffer) == 0:
-                    raise socket.error("Connection closed by remote server.")
+                if HIREDIS_USE_BYTE_BUFFER:
+                    bufflen = recv_into(self._sock, self._buffer)
+                    if bufflen == 0:
+                        raise socket.error(SERVER_CLOSED_CONNECTION_ERROR)
+                else:
+                    buffer = recv(self._sock, socket_read_size)
+                    # an empty string indicates the server shutdown the socket
+                    if not isinstance(buffer, bytes) or len(buffer) == 0:
+                        raise socket.error(SERVER_CLOSED_CONNECTION_ERROR)
             except socket.timeout:
                 raise TimeoutError("Timeout reading from socket")
             except socket.error:
                 e = sys.exc_info()[1]
                 raise ConnectionError("Error while reading from socket: %s" %
                                       (e.args,))
-            if not buffer:
-                raise ConnectionError("Socket closed on remote end")
-            self._reader.feed(buffer)
-            # proactively, but not conclusively, check if more data is in the
-            # buffer. if the data received doesn't end with \r\n, there's more.
-            if not buffer.endswith(SYM_CRLF):
-                continue
+            if HIREDIS_USE_BYTE_BUFFER:
+                self._reader.feed(self._buffer, 0, bufflen)
+            else:
+                self._reader.feed(buffer)
             response = self._reader.gets()
         # if an older version of hiredis is installed, we need to attempt
         # to convert ResponseErrors to their appropriate types.
@@ -348,6 +424,7 @@ class HiredisParser(BaseParser):
             raise response[0]
         return response
 
+
 if HIREDIS_AVAILABLE:
     DefaultParser = HiredisParser
 else:
@@ -374,9 +451,7 @@ class Connection(object):
         self.socket_keepalive = socket_keepalive
         self.socket_keepalive_options = socket_keepalive_options or {}
         self.retry_on_timeout = retry_on_timeout
-        self.encoding = encoding
-        self.encoding_errors = encoding_errors
-        self.decode_responses = decode_responses
+        self.encoder = Encoder(encoding, encoding_errors, decode_responses)
         self._sock = None
         self._parser = parser_class(socket_read_size=socket_read_size)
         self._description_args = {
@@ -407,6 +482,8 @@ class Connection(object):
             return
         try:
             sock = self._connect()
+        except socket.timeout:
+            raise TimeoutError("Timeout connecting to server")
         except socket.error:
             e = sys.exc_info()[1]
             raise ConnectionError(self._error_message(e))
@@ -518,11 +595,12 @@ class Connection(object):
             e = sys.exc_info()[1]
             self.disconnect()
             if len(e.args) == 1:
-                _errno, errmsg = 'UNKNOWN', e.args[0]
+                errno, errmsg = 'UNKNOWN', e.args[0]
             else:
-                _errno, errmsg = e.args
+                errno = e.args[0]
+                errmsg = e.args[1]
             raise ConnectionError("Error %s while writing to socket. %s." %
-                                  (_errno, errmsg))
+                                  (errno, errmsg))
         except:
             self.disconnect()
             raise
@@ -531,13 +609,14 @@ class Connection(object):
         "Pack and send a command to the Redis server"
         self.send_packed_command(self.pack_command(*args))
 
-    def can_read(self):
+    def can_read(self, timeout=0):
         "Poll the socket to see if there's data that can be read."
         sock = self._sock
         if not sock:
             self.connect()
             sock = self._sock
-        return bool(select([sock], [], [], 0)[0]) or self._parser.can_read()
+        return self._parser.can_read() or \
+            bool(select([sock], [], [], timeout)[0])
 
     def read_response(self):
         "Read the response from a previously sent command"
@@ -550,22 +629,6 @@ class Connection(object):
             raise response
         return response
 
-    def encode(self, value):
-        "Return a bytestring representation of the value"
-        if isinstance(value, Token):
-            return b(value.value)
-        elif isinstance(value, bytes):
-            return value
-        elif isinstance(value, (int, long)):
-            value = b(str(value))
-        elif isinstance(value, float):
-            value = b(repr(value))
-        elif not isinstance(value, basestring):
-            value = str(value)
-        if isinstance(value, unicode):
-            value = value.encode(self.encoding, self.encoding_errors)
-        return value
-
     def pack_command(self, *args):
         "Pack a series of arguments into the Redis protocol"
         output = []
@@ -576,14 +639,15 @@ class Connection(object):
         # to prevent them from being encoded.
         command = args[0]
         if ' ' in command:
-            args = tuple([Token(s) for s in command.split(' ')]) + args[1:]
+            args = tuple([Token.get_token(s)
+                          for s in command.split()]) + args[1:]
         else:
-            args = (Token(command),) + args[1:]
+            args = (Token.get_token(command),) + args[1:]
 
         buff = SYM_EMPTY.join(
             (SYM_STAR, b(str(len(args))), SYM_CRLF))
 
-        for arg in imap(self.encode, args):
+        for arg in imap(self.encoder.encode, args):
             # to avoid large string mallocs, chunk the command into the
             # output list if we're sending large values
             if len(buff) > 6000 or len(arg) > 6000:
@@ -598,6 +662,26 @@ class Connection(object):
         output.append(buff)
         return output
 
+    def pack_commands(self, commands):
+        "Pack multiple commands into the Redis protocol"
+        output = []
+        pieces = []
+        buffer_length = 0
+
+        for cmd in commands:
+            for chunk in self.pack_command(*cmd):
+                pieces.append(chunk)
+                buffer_length += len(chunk)
+
+            if buffer_length > 6000:
+                output.append(SYM_EMPTY.join(pieces))
+                buffer_length = 0
+                pieces = []
+
+        if pieces:
+            output.append(SYM_EMPTY.join(pieces))
+        return output
+
 
 class SSLConnection(Connection):
     description_format = "SSLConnection<host=%(host)s,port=%(port)s,db=%(db)s>"
@@ -652,9 +736,7 @@ class UnixDomainSocketConnection(Connection):
         self.password = password
         self.socket_timeout = socket_timeout
         self.retry_on_timeout = retry_on_timeout
-        self.encoding = encoding
-        self.encoding_errors = encoding_errors
-        self.decode_responses = decode_responses
+        self.encoder = Encoder(encoding, encoding_errors, decode_responses)
         self._sock = None
         self._parser = parser_class(socket_read_size=socket_read_size)
         self._description_args = {
@@ -681,10 +763,29 @@ class UnixDomainSocketConnection(Connection):
                 (exception.args[0], self.path, exception.args[1])
 
 
+FALSE_STRINGS = ('0', 'F', 'FALSE', 'N', 'NO')
+
+
+def to_bool(value):
+    if value is None or value == '':
+        return None
+    if isinstance(value, basestring) and value.upper() in FALSE_STRINGS:
+        return False
+    return bool(value)
+
+
+URL_QUERY_ARGUMENT_PARSERS = {
+    'socket_timeout': float,
+    'socket_connect_timeout': float,
+    'socket_keepalive': to_bool,
+    'retry_on_timeout': to_bool
+}
+
+
 class ConnectionPool(object):
     "Generic connection pool"
     @classmethod
-    def from_url(cls, url, db=None, **kwargs):
+    def from_url(cls, url, db=None, decode_components=False, **kwargs):
         """
         Return a connection pool configured from the given URL.
 
@@ -695,9 +796,14 @@ class ConnectionPool(object):
             unix://[:password]@/path/to/socket.sock?db=0
 
         Three URL schemes are supported:
-            redis:// creates a normal TCP socket connection
-            rediss:// creates a SSL wrapped TCP socket connection
-            unix:// creates a Unix Domain Socket connection
+
+        - ```redis://``
+          <http://www.iana.org/assignments/uri-schemes/prov/redis>`_ creates a
+          normal TCP socket connection
+        - ```rediss://``
+          <http://www.iana.org/assignments/uri-schemes/prov/rediss>`_ creates a
+          SSL wrapped TCP socket connection
+        - ``unix://`` creates a Unix Domain Socket connection
 
         There are several ways to specify a database number. The parse function
         will return the first specified option:
@@ -708,9 +814,20 @@ class ConnectionPool(object):
 
         If none of these options are specified, db=0 is used.
 
+        The ``decode_components`` argument allows this function to work with
+        percent-encoded URLs. If this argument is set to ``True`` all ``%xx``
+        escapes will be replaced by their single-character equivalents after
+        the URL has been parsed. This only applies to the ``hostname``,
+        ``path``, and ``password`` components.
+
         Any additional querystring arguments and keyword arguments will be
-        passed along to the ConnectionPool class's initializer. In the case
-        of conflicting arguments, querystring arguments always win.
+        passed along to the ConnectionPool class's initializer. The querystring
+        arguments ``socket_connect_timeout`` and ``socket_timeout`` if supplied
+        are parsed as float values. The arguments ``socket_keepalive`` and
+        ``retry_on_timeout`` are parsed to boolean values that accept
+        True/False, Yes/No values to indicate state. Invalid types cause a
+        ``UserWarning`` to be raised. In the case of conflicting arguments,
+        querystring arguments always win.
         """
         url_string = url
         url = urlparse(url)
@@ -730,28 +847,46 @@ class ConnectionPool(object):
 
         for name, value in iteritems(parse_qs(qs)):
             if value and len(value) > 0:
-                url_options[name] = value[0]
+                parser = URL_QUERY_ARGUMENT_PARSERS.get(name)
+                if parser:
+                    try:
+                        url_options[name] = parser(value[0])
+                    except (TypeError, ValueError):
+                        warnings.warn(UserWarning(
+                            "Invalid value for `%s` in connection URL." % name
+                        ))
+                else:
+                    url_options[name] = value[0]
+
+        if decode_components:
+            password = unquote(url.password) if url.password else None
+            path = unquote(url.path) if url.path else None
+            hostname = unquote(url.hostname) if url.hostname else None
+        else:
+            password = url.password
+            path = url.path
+            hostname = url.hostname
 
         # We only support redis:// and unix:// schemes.
         if url.scheme == 'unix':
             url_options.update({
-                'password': url.password,
-                'path': url.path,
+                'password': password,
+                'path': path,
                 'connection_class': UnixDomainSocketConnection,
             })
 
         else:
             url_options.update({
-                'host': url.hostname,
+                'host': hostname,
                 'port': int(url.port or 6379),
-                'password': url.password,
+                'password': password,
             })
 
             # If there's a path argument, use it as the db argument if a
             # querystring value wasn't specified
-            if 'db' not in url_options and url.path:
+            if 'db' not in url_options and path:
                 try:
-                    url_options['db'] = int(url.path.replace('/', ''))
+                    url_options['db'] = int(path.replace('/', ''))
                 except (AttributeError, ValueError):
                     pass
 
@@ -763,6 +898,17 @@ class ConnectionPool(object):
 
         # update the arguments from the URL values
         kwargs.update(url_options)
+
+        # backwards compatability
+        if 'charset' in kwargs:
+            warnings.warn(DeprecationWarning(
+                '"charset" is deprecated. Use "encoding" instead'))
+            kwargs['encoding'] = kwargs.pop('charset')
+        if 'errors' in kwargs:
+            warnings.warn(DeprecationWarning(
+                '"errors" is deprecated. Use "encoding_errors" instead'))
+            kwargs['encoding_errors'] = kwargs.pop('errors')
+
         return cls(**kwargs)
 
     def __init__(self, connection_class=Connection, max_connections=None,
@@ -771,8 +917,8 @@ class ConnectionPool(object):
         Create a connection pool. If max_connections is set, then this
         object raises redis.ConnectionError when the pool's limit is reached.
 
-        By default, TCP connections are created connection_class is specified.
-        Use redis.UnixDomainSocketConnection for unix sockets.
+        By default, TCP connections are created unless connection_class is
+        specified. Use redis.UnixDomainSocketConnection for unix sockets.
 
         Any additional keyword arguments are passed to the constructor of
         connection_class.
@@ -820,6 +966,15 @@ class ConnectionPool(object):
         self._in_use_connections.add(connection)
         return connection
 
+    def get_encoder(self):
+        "Return an encoder based on encoding settings"
+        kwargs = self.connection_kwargs
+        return Encoder(
+            encoding=kwargs.get('encoding', 'utf-8'),
+            encoding_errors=kwargs.get('encoding_errors', 'strict'),
+            decode_responses=kwargs.get('decode_responses', False)
+        )
+
     def make_connection(self):
         "Create a new connection"
         if self._created_connections >= self.max_connections:
